/*    $Id: ExprDouble.java,v 1.1 1996/02/27 04:17:38 davidliu Exp $
**
**
**************************************************************************
**    Copyright (C) 1996 David Wanqian Liu
**                       davidliu@mti.sgi.com
**                       http://reality.sgi.com/exmployees/davidliu_mti
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**************************************************************************
*/


package expression;



import java.lang.*;


/**
  A class defined to output more flexible double representation than
  the standard java.lang.Double.
  */
public class ExprDouble {
  /**
    Convert a double into a string using default precision 15.
    */
  public static String toString(double val) {
    return toString(val, 15);
  }


  /**
    Convert a double into a string using specific precision.
    */
  public static String toString(double val, int precision) {

    if (Double.isNaN(val) || Double.isInfinite(val))
      return Double.toString(val);
    
    String origStr = String.valueOf(val);

    // find out the exponent
    int expnPos = origStr.lastIndexOf('e');
    
    if ((expnPos > 0) &&
	((Math.abs(val) > Long.MAX_VALUE) ||
	(Math.abs(val) < Math.pow(10.0, -precision)))) {
      String expnStr = origStr.substring(expnPos+1);

      double base =
	val / Math.pow(10.0, Double.valueOf(expnStr).doubleValue());

      return baseToString(base, precision) + "e" + expnStr;
    }
    else {
      return baseToString(val, precision);
    }
  }


  /**
    Convert a base double value (no exponential component) into a string
    using specific precision.
    */
  private static String baseToString(double val, int precision) {
    double absval = Math.abs(val);

    String signStr = "";
    if (val < 0) signStr = "-";

    long intDigit = (long) Math.floor(absval);
    String intDigitStr = String.valueOf(intDigit);

    // if the number is very large, double can't provide enough precision
    precision -= intDigitStr.length();

    String precDigitStr = "";
    if (precision > 0) {
      long precDigit =
	Math.round((absval - intDigit) * Math.pow(10.0, precision));
      precDigitStr = String.valueOf(precDigit);

      // Find out number of 0's right after the precision point
      String zeroFilling = "";
      for(int i=0; i<precision-precDigitStr.length(); i++)
	zeroFilling += "0";
      precDigitStr = zeroFilling + precDigitStr;
      
      // Get rid of trailing 0's, find the last 0
      int lastZero;
      for (lastZero=precDigitStr.length()-1; lastZero>=0; lastZero--)
	if (precDigitStr.charAt(lastZero) != '0') break;
      
      precDigitStr = precDigitStr.substring(0, lastZero+1);
    }

    if (precDigitStr.equals(""))
      return signStr + intDigitStr;
    else
      return signStr + intDigitStr + "." + precDigitStr;
  }
}
