package graphoutput;

import javax.media.j3d.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.behaviors.picking.*;
import javax.vecmath.*;
import java.util.*;
import java.awt.*;

/**
 * This class contains the implementation of the method
 * "Linking With Wings". It creates the geometry within
 * this BranchGroup into the 1x1x1-box.
 */
public class LWWGraph extends BranchGroup
{                      
    public final static float BASE_LINE_THICKNESS = 3.0f;
    public final static float WING_LINE_THICKNESS = 1.0f;
    private int dimcnt, pointcnt;

	private Style style;
	private GraphOutput parent;
    private BranchGroup stripe_bg = new BranchGroup();

    /**
     * Removes the current geometry (stripes)
     */
    public void removeGeometry()
    {
      stripe_bg.detach();
    }

    /**
     * Draws new stripes using the specified data
     */
    public void updateGeometry(Vector points_v, Vector scaling_v)
    {
      stripe_bg = new BranchGroup();
      stripe_bg.setCapability(BranchGroup.ALLOW_DETACH);

      createGeometries(dimcnt, pointcnt, points_v, scaling_v);

      addChild(stripe_bg);

      style.resetMarkers();
    }

    /**
     * Fills the stripeShape_v with the stripes as Shape3Ds
     */
	private void createGeometries(int dimcnt, int pointcnt, Vector points_v, Vector scaling_v)
	{
	  LineStripArray line;
	  Shape3D lineShape = new Shape3D();
	  Vector wingArray_v = new Vector();
	  Vector wingShape_v = new Vector();
	  Vector planeArray_v = new Vector();
	  Vector planeShape_v = new Vector();
      Vector backArray_v = new Vector();
      Vector backShape_v = new Vector();
      Vector3d v = new Vector3d(1.0, 1.0, 1.0);
	  int[] linelength;
	  linelength = new int[1];
      linelength[0]=pointcnt;
      int[] planelength = { pointcnt*2 };

      // hide the coordinate-box
      parent.hideCoordinateBox();

      // first draw a plane on the bottom side of the coordbox
      Appearance ground_app = new Appearance();
      Color3f objColor = new Color3f(0.2f, 0.2f, 0.2f);
      Material mat = new Material(objColor, new Color3f(0.0f,0.0f,0.0f), objColor,
	                                              objColor, 20.0f);
      ground_app.setMaterial(mat);
      TransformGroup ground_tg = new TransformGroup();
       Transform3D ground_t3d = new Transform3D();
       ground_t3d.setTranslation(new Vector3d(0.5, 0.5, 0.5));
       ground_tg.setTransform(ground_t3d);
      ground_tg.addChild(new Box(0.5f, 0.005f, 0.5f, ground_app));
      stripe_bg.addChild(ground_tg);

      // draw the trajectory
	  line = new LineStripArray(pointcnt, GeometryArray.COORDINATES, linelength);
	  for(int d=2; d<dimcnt; d++)
	  {
	    LineStripArray lsa = new LineStripArray(pointcnt, GeometryArray.COORDINATES, linelength);
        wingArray_v.addElement(lsa);
        TriangleStripArray tsa = new TriangleStripArray(2*pointcnt,
                                                        GeometryArray.COORDINATES | GeometryArray.NORMALS,
                                                        planelength);
        planeArray_v.addElement(tsa);
        TriangleStripArray tsa1 = new TriangleStripArray(2*pointcnt,
                                                        GeometryArray.COORDINATES | GeometryArray.NORMALS,
                                                        planelength);
        backArray_v.addElement(tsa1);
	  }

	  for (int i=0; i<pointcnt; i++)
	  {
	    // baseline
	    double x = ((Range)scaling_v.elementAt(0)).normalize(
		               ((Double)((Vector)points_v.elementAt(i)).elementAt(0)).doubleValue());
		double y = 0.51;
	    double z = ((Range)scaling_v.elementAt(1)).normalize(
		               ((Double)((Vector)points_v.elementAt(i)).elementAt(1)).doubleValue());
	    line.setCoordinate(i, new Point3d(x, y, z));
	    //--

	    // wing-lines
        if(i<pointcnt-1)
        {
          double vx = ((Range)scaling_v.elementAt(0)).normalize(
		              ((Double)((Vector)points_v.elementAt(i+1)).elementAt(0)).doubleValue()) - x;
          double vy = 0.0;
          double vz = ((Range)scaling_v.elementAt(1)).normalize(
		              ((Double)((Vector)points_v.elementAt(i+1)).elementAt(1)).doubleValue()) - z;
		  v = new Vector3d(vx, vy, vz);
		  v.normalize();
		}

	    for(int d=2; d<dimcnt; d++)
	    {
          double angle = (d-1) * Math.PI/(double)(dimcnt-1);
          Vector3d p = new Vector3d(v.z*Math.cos(angle),
                                    Math.sin(angle),
                                    -v.x*Math.cos(angle));
          p.scale(parent.style.maxwinglength);
          p.scale(((Range)scaling_v.elementAt(d)).normalize(
		               ((Double)((Vector)points_v.elementAt(i)).elementAt(d)).doubleValue()) );
          Point3d p1 = new Point3d(x+p.x, y+p.y, z+p.z);

          ((LineStripArray)wingArray_v.elementAt(d-2)).setCoordinate(i, p1);

          // wing-planes
		  double[] a = { x, y, z, p1.x, p1.y, p1.z };
          ((TriangleStripArray)planeArray_v.elementAt(d-2)).setCoordinates(i*2, a);
          ((TriangleStripArray)backArray_v.elementAt(d-2)).setCoordinates(i*2, a);
          if(i<pointcnt-1) // calc normal, leave the last dataset
          {
            Vector3d n1 = new Vector3d();
            n1.cross(v, p);
            n1.normalize();
            Vector3f n = new Vector3f(n1);
            ((TriangleStripArray)planeArray_v.elementAt(d-2)).setNormal(i*2, n);
            ((TriangleStripArray)planeArray_v.elementAt(d-2)).setNormal(i*2+1, n);
            ((TriangleStripArray)backArray_v.elementAt(d-2)).setNormal(i*2, n);
            ((TriangleStripArray)backArray_v.elementAt(d-2)).setNormal(i*2+1, n);
          }
          //--

	    }
	    //--
	  }

	  lineShape.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
	  lineShape.setGeometry(line);
	  for(int d=2; d<dimcnt; d++)
	  {
	    Shape3D s3d = new Shape3D();
	    s3d.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
	    s3d.setGeometry( (LineStripArray)wingArray_v.elementAt(d-2) );
	    wingShape_v.addElement(s3d);
	    Shape3D s3d1 = new Shape3D();
	    s3d1.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
	    s3d1.setGeometry ( (TriangleStripArray)planeArray_v.elementAt(d-2) );
	    planeShape_v.addElement(s3d1);
        Shape3D s3d2 = new Shape3D();
	    s3d2.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
        s3d2.setGeometry ( (TriangleStripArray)backArray_v.elementAt(d-2) );
        backShape_v.addElement(s3d2);
      }

      // create the lineappearance from the stored material-appearance
      Appearance temp_app = style.getAppearance(0);
      Material temp_mat = temp_app.getMaterial();
      Color3f temp_col = new Color3f();
      temp_mat.getAmbientColor(temp_col);

	  Appearance app = new Appearance();
	  Color3f color = temp_col;
      ColoringAttributes ca = new ColoringAttributes();
      ca.setColor(color);
      app.setColoringAttributes(ca);

      LineAttributes lat = new LineAttributes();
      lat.setLineWidth(BASE_LINE_THICKNESS);
//?      lat.setLineAntialiasingEnable(true);
      app.setLineAttributes(lat);

      lineShape.setAppearance(app);

      for (int i=2; i<dimcnt; i++)
      {
        temp_app = style.getAppearance(i);
        temp_mat = temp_app.getMaterial();
        temp_col = new Color3f();
        temp_mat.getAmbientColor(temp_col);

	    app = new Appearance();
	    color = temp_col;
        ca = new ColoringAttributes();
        ca.setColor(color);
        app.setColoringAttributes(ca);

        lat = new LineAttributes();
        lat.setLineWidth(WING_LINE_THICKNESS);
//?        lat.setLineAntialiasingEnable(true);
        app.setLineAttributes(lat);

        ((Shape3D)wingShape_v.elementAt(i-2)).setAppearance(app);

        Appearance plane_app = new Appearance();
        plane_app.setMaterial(temp_mat);
        PolygonAttributes pa = new PolygonAttributes();
        pa.setCullFace(pa.CULL_FRONT);
        plane_app.setPolygonAttributes(pa);

        TransparencyAttributes ta = new TransparencyAttributes();
        ta.setTransparencyMode(ta.BLENDED);
        ta.setTransparency(parent.style.wingtransparency);
        ta.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
        plane_app.setTransparencyAttributes(ta);

        ((Shape3D)planeShape_v.elementAt(i-2)).setAppearance(plane_app);

        Appearance back_app = new Appearance();
        back_app.setMaterial(temp_mat);
        PolygonAttributes pa1 = new PolygonAttributes();
        pa1.setCullFace(pa1.CULL_BACK);
        back_app.setPolygonAttributes(pa1);

        TransparencyAttributes ta1 = new TransparencyAttributes();
        ta1.setTransparencyMode(ta1.BLENDED);
        ta1.setTransparency(parent.style.wingtransparency);
        ta1.setCapability(TransparencyAttributes.ALLOW_VALUE_READ);
        back_app.setTransparencyAttributes(ta1);

        ((Shape3D)backShape_v.elementAt(i-2)).setAppearance(back_app);

      }

	  stripe_bg.addChild(lineShape);
	  for (int d=2; d<dimcnt; d++)
	  {
	    stripe_bg.addChild( (Shape3D)wingShape_v.elementAt(d-2) );
        stripe_bg.addChild( (Shape3D)planeShape_v.elementAt(d-2) );
        stripe_bg.addChild( (Shape3D)backShape_v.elementAt(d-2) );
	  }

	}

    /**
     * Draws the ranges of each coordinate into the scene
     */
    private void createAnnotation(int dimcnt, Vector scaling_v)
	{
	  TransformGroup tg, tga;
	  Transform3D t;
      Text2D txt;
      Vector3d[] pos = { new Vector3d(0.0, 0.03, 0.5),
                         new Vector3d(0.9, 0.03, 0.5),
                         new Vector3d(-0.15, -0.03, 0.5),
                         new Vector3d(-0.15, -1.0, 0.5) };
      double[] val = { ((Range)scaling_v.elementAt(0)).getMin2(),
                       ((Range)scaling_v.elementAt(0)).getMax2(),
                       ((Range)scaling_v.elementAt(1)).getMin2(),
                       ((Range)scaling_v.elementAt(1)).getMax2() };


      tga = new TransformGroup();

      for(int i=0; i<4; i++)
      {
	    tg = new TransformGroup();
	    t = new Transform3D();
	    t.setTranslation( pos[i] );
	    tg.setTransform(t);
	    txt = new Text2D(""+val[i], new Color3f(0.9f, 1.0f, 1.0f), "System", 15, 1);
        tg.addChild(txt);
	    tga.addChild(tg);
	  }
	  for(int d=2; d<dimcnt; d++)
      {
	    tg = new TransformGroup();
	    t = new Transform3D();
	    t.setTranslation(new Vector3d(0.1, -1.1-(double)(d-2)*0.1, 0.5));
	    tg.setTransform(t);
	    txt = new Text2D("["+((Range)scaling_v.elementAt(d)).getMin2()+
	                     "; "+((Range)scaling_v.elementAt(d)).getMax2()+"]",
	                     new Color3f(0.9f, 1.0f, 1.0f), "System", 15, 1);
        tg.addChild(txt);
	    tga.addChild(tg);
      }

	  t = new Transform3D();
	  t.rotX(-Math.PI/2.0);
	  tga.setTransform(t);
	  addChild(tga);

	}

   /**
    * Inserts text3D-elements into the scene specifying the letters
    * for each variable.
    */
    private void createVariables(int dimcnt, Vector variablenames_v)
	{

	  TransformGroup tg, tga;
	  Transform3D t;
      Text2D txt;

      tga = new TransformGroup();

      tg = new TransformGroup();
	  t = new Transform3D();
	  t.setTranslation( new Vector3d(0.47, 0.03, 0.5) );
	  tg.setTransform(t);
	  txt = new Text2D(""+((String)variablenames_v.elementAt(0)), new Color3f(0.7f, 0.8f, 1.0f), "System", 18, 1);
      tg.addChild(txt);
	  tga.addChild(tg);

      tg = new TransformGroup();
	  t = new Transform3D();
	  t.setTranslation( new Vector3d(-0.1, -0.53, 0.5) );
	  tg.setTransform(t);
	  txt = new Text2D(""+((String)variablenames_v.elementAt(1)), new Color3f(0.7f, 0.8f, 1.0f), "System", 18, 1);
      tg.addChild(txt);
	  tga.addChild(tg);

	  for(int i=2; i<dimcnt; i++)
	  {
        tg = new TransformGroup();
  	    t = new Transform3D();
	    t.setTranslation( new Vector3d(0.0, -1.1-(i-2)*0.1, 0.5) );
	    tg.setTransform(t);
        Color3f col = new Color3f();
	    parent.style.getAppearance(i).getMaterial().getAmbientColor(col);
	    txt = new Text2D(""+((String)variablenames_v.elementAt(i)), col, "System", 18, 1);
        tg.addChild(txt);
	    tga.addChild(tg);
	  }

	  t = new Transform3D();
	  t.rotX(-Math.PI/2.0);
	  tga.setTransform(t);
	  addChild(tga);

	}

   /**
    * Inserts the spheres, which are dragable with the mouse to change
    * the starting value of each variable
    */
    private void createController(Vector points_v, Vector scaling_v)
    {
      Appearance app = new Appearance();
      Color3f objColor = new Color3f(1.0f, 0.2f, 0.0f);

      Material mat = new Material(objColor, new Color3f(0.0f,0.0f,0.0f), objColor,
	                                              new Color3f(1.0f,1.0f,1.0f), 80.0f);
      app.setMaterial(mat);


      BranchGroup bg = new BranchGroup();
      addChild(bg);

      TransformGroup tg = new TransformGroup();
      Transform3D t3d = new Transform3D();

	  double x = ((Range)scaling_v.elementAt(0)).normalize(
		               ((Double)((Vector)points_v.elementAt(0)).elementAt(0)).doubleValue());
	  double y = ((Range)scaling_v.elementAt(1)).normalize(
		               ((Double)((Vector)points_v.elementAt(0)).elementAt(1)).doubleValue());

      t3d.setTranslation(new Vector3d(x, 0.507, y));
      tg.setTransform(t3d);
      tg.setCapability(Node.ENABLE_PICK_REPORTING);
      tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      bg.addChild(tg);

      tg.addChild((new Sphere(0.015f, app)));

      PickTranslateBehavior3 mt = new PickTranslateBehavior3(bg,
                                                parent.canvas,
                                                new BoundingSphere(new Point3d(0.0, 1.0, 1.0), 0.015),
                                                PickObject.USE_BOUNDS,
                                                parent);
      bg.addChild(mt);


   }



   /**
    * By calling this constructor the graph is created using the specified
    * data.
    */
    public LWWGraph(GraphOutput parent, int dimcnt, int pointcnt,
	                Vector points_v, Vector scaling_v, Vector variablenames_v,
					Style style)
	{
	  this.parent = parent;
	  this.style = style;
	  this.dimcnt=dimcnt;
	  this.pointcnt=pointcnt;

      stripe_bg.setCapability(BranchGroup.ALLOW_DETACH);
      addChild(stripe_bg);

	  createGeometries(dimcnt, pointcnt, points_v, scaling_v);
      createVariables(dimcnt, variablenames_v);
      createController(points_v, scaling_v);
      if(style.showscaling==true) createAnnotation(dimcnt, scaling_v);

      setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
      setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);

	  setCapability(BranchGroup.ALLOW_DETACH);
	  compile();
	}

} // end of class LWWGraph
