/*
  draws a polygonal line representing a point in the
  trajectory of TDC
 */

package graphoutput;

import java.awt.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import java.util.*;
import com.sun.j3d.utils.geometry.*;

public class TDCHandle extends BranchGroup {
//   private Appearance app;
//   private Shape3D shape;
//   private TransformGroup tg;

   private static final int LINE_THICKNESS = 3;
   private static final Color3f DEFAULT_COLOR =
           new Color3f(1.0f, 1.0f, 0.7f);
   private static final float DEFAULT_TRANSPARENCY = 0.0f;
   private int dimcnt;
   private Color3f color;
   private float transparency;
   private BranchGroup marker_bg;
   private double radius;
   private int vertexcnt;
   private Vector aktpos_v;

   LineStripArray lsa;
   Vector tg_v = new Vector();

   /**
    * create an empty branchgroup
    */
   public TDCHandle(double radius, int vertexcnt, int dimcnt)
   {
    this.dimcnt=dimcnt;
    this.radius=radius;
    this.vertexcnt=vertexcnt;
    this.color=DEFAULT_COLOR;
    this.transparency=DEFAULT_TRANSPARENCY;
    marker_bg = new BranchGroup();
    setColor(color, transparency);

    setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    setCapability(BranchGroup.ALLOW_DETACH);

    marker_bg.setCapability(BranchGroup.ALLOW_DETACH);
    addChild(marker_bg);
   }

  /**
   * creates an instance of Appearance with the specified
   * attributes
   */
  private Appearance createAppearance(Color3f color, float transparency)
  {
    Appearance app = new Appearance();

	Material mat = new Material(color, new Color3f(0.0f,0.0f,0.0f), color,
                                  new Color3f(1.0f,1.0f,1.0f), 80.0f);
    app.setMaterial(mat);

    TransparencyAttributes ta = new TransparencyAttributes();
    ta.setTransparency(transparency);
    app.setTransparencyAttributes(ta);

    return app;
  }

  /**
   * creates an instance of an appearance for a line with the specified
   * attributs
   */
  private Appearance createLineAppearance(Color3f color, float transparency)
  {
	Appearance app = new Appearance();

    ColoringAttributes ca = new ColoringAttributes();
    ca.setColor(color);
    app.setColoringAttributes(ca);

    LineAttributes lat = new LineAttributes();
    lat.setLineWidth(LINE_THICKNESS);
    lat.setLineAntialiasingEnable(true);
    app.setLineAttributes(lat);
    return app;
  }

  /**
   * changes the current position and orientation of the plane
   */
  public void setPosition(Vector pos_v)
  {
    if(lsa!=null)  // the marker is already present in the scene
    {
      aktpos_v = (Vector)pos_v.clone();

      int plane_cnt = (1+dimcnt)/2;
      double space = 1.0;
      if(plane_cnt>1)
        space = 1 / (double)(plane_cnt-1);
      else space = 1.0;
      for(int i=0; i<plane_cnt; i++)
      {
        int f=1;
        if (2*i+1==dimcnt) f=0;
        Point3d p = new Point3d(  ((Double)pos_v.elementAt(2*i)).doubleValue(),
                                  1.0-(double)i*space,
                                  ((Double)pos_v.elementAt(2*i+f)).doubleValue() );
        if(i==0) lsa.setCoordinate(0, new Point3d(p.x, MarkerTDC.SPHERE_HEIGHT, p.z));
        lsa.setCoordinate(i+1, p);
        Transform3D t3d = new Transform3D();
        t3d.setTranslation(new Vector3d(p));
        ((TransformGroup)tg_v.elementAt(i)).setTransform(t3d);
      }
    }
    else   // the marker has to be drawn
    {
      BranchGroup new_bg = createMarker(pos_v);

      marker_bg.detach();

      marker_bg=new BranchGroup();
      marker_bg=new_bg;
      addChild(marker_bg);
    }
/*
    BranchGroup new_bg = createMarker(pos_v);

    marker_bg.detach();

    marker_bg=new BranchGroup();
    marker_bg=new_bg;
    addChild(marker_bg);
*/
  }

  /**
   *  set the color and the transparency of the TDCHandle
   */
  public void setColor(Color3f col, float transparency)
  {
    this.color = col;
    this.transparency = transparency;

    if(aktpos_v!=null)
    {
      BranchGroup new_bg = createMarker(aktpos_v);
      marker_bg.detach();
      marker_bg=new BranchGroup();
      marker_bg=new_bg;
      addChild(marker_bg);
    }
  }

  /**
   *  returns the current color of the TDCHandle
   */
  public Color3f getColor()
  {
    return color;
  }

  /**
   *  returns the current transparency of the TDCHandle
   */
  public float getTransparency()
  {
    return transparency;
  }

  /**
   *  creates the marking polygon with the specified
   *  coordinates
   */
  public BranchGroup createMarker(Vector pos_v)
  {
	BranchGroup bg = new BranchGroup();
    bg.setCapability(BranchGroup.ALLOW_DETACH);

	Appearance app = createAppearance(color, transparency);
    Appearance app1 = createLineAppearance(color, transparency);

    // insert the geometry

	int plane_cnt = (1+dimcnt)/2;
    double space = 1.0;
    if(plane_cnt>1)
      space = 1 / (double)(plane_cnt-1);
    else space = 1.0;

    int[] ll = { plane_cnt+1 };
    lsa =
	   new LineStripArray(plane_cnt+1, GeometryArray.COORDINATES, ll);
    lsa.setCapability(GeometryArray.ALLOW_COORDINATE_WRITE);

    tg_v = new Vector();
    for(int i=0; i<plane_cnt; i++)
    {
	  int f=1;
      if (2*i+1==dimcnt) f=0;
      TransformGroup tg = new TransformGroup();
      tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
       Transform3D t3d = new Transform3D();
       Point3d p = new Point3d(  ((Double)pos_v.elementAt(2*i)).doubleValue(),
                                 1.0-(double)i*space,
                                 ((Double)pos_v.elementAt(2*i+f)).doubleValue() );
       Vector3d vp = new Vector3d(p);
       t3d.setTranslation(vp);
       if(i==0) lsa.setCoordinate(0, new Point3d(p.x, MarkerTDC.SPHERE_HEIGHT, p.z));
       lsa.setCoordinate(i+1, p);
       tg.setTransform(t3d);
      tg_v.addElement(tg);
      tg.addChild(new Sphere((float)radius, app));
      bg.addChild(tg);
    }

    Shape3D line = new Shape3D();
    line.setGeometry(lsa);
    line.setAppearance(app1);
    bg.addChild(line);


    return bg;

  }


}